;; Generated package description from dired-du.el  -*- no-byte-compile: t -*-
(define-package "dired-du" "0.5.2.0.20221221.75108" "Dired with recursive directory sizes" '((emacs "24.4") (cl-lib "0.5")) :commit "f7e1593e94388b0dfb71af8e9a3d5d07edf5a159" :url "https://elpa.gnu.org/packages/dired-du.html" :authors '(("Tino Calancha" . "tino.calancha@gmail.com")) :maintainer '("Tino Calancha" . "tino.calancha@gmail.com") :keywords '("files" "unix" "convenience"))
